#ifndef ARDUINO_CONFIGEEPROM_H
#define ARDUINO_CONFIGEEPROM_H

#include "SimpleTimeout.h"
#include "PowerBandMap.h"
/**
 * @author James Michener (c) 2012
 *
 * an area of RAM that can be read at start up,  and can be changed
 * then periodically checks to see if anything has changed, and if so
 * it will store the data back to EEPROM,  saving configurations
 */
/// must be defined elsewhere in the project and before this INCLUDE file!!
const byte maxConfigEEPromByte = maxNumberPowerBand + 10;  // how many bytes are we dealing with maximum
const unsigned int offsetInEEPromConfig = 0;  /// where in the EEPROM is this being stored.
const unsigned long timeOutToWriteEEPROMData = 10000; // number of ms between writing data to EEPROM

class configEEProm
{
public:
	configEEProm(void) { st.start(timeOutToWriteEEPROMData ); }
	~configEEProm(void) {};

	/// accessing each bit of data using the byte and bit offset
	inline bool bitReadEE(const byte byteOffset, const byte bitOffset) const { return bitRead(data[byteOffset],bitOffset); }
	inline void bitSetEE(const byte byteOffset, const byte bitOffset) { bitSet(data[byteOffset],bitOffset); }
	inline void bitClearEE(const byte byteOffset, const byte bitOffset) { bitClear(data[byteOffset],bitOffset); }

	/// accessing each bit of data using a single byte where the 3 lsb bits are the bit position
	/// and the 5 msb (32 max) bytes are the offset.
	inline bool bitReadEE(const byte bitID) const { return bitReadEE(bitID >> 3, bitID & 0x07); }
	inline void bitSetEE(const byte bitID) { bitSetEE(bitID >> 3, bitID & 0x07); }
	inline void bitClearEE(const byte bitID) { bitClearEE(bitID >> 3 , bitID & 0x07); }

    inline void byteSetEE(const byte byteOffset) { data[byteOffset] = 0xff; }
    inline void byteClearEE(const byte byteOffset) { data[byteOffset] = 0; }
	/// used at start up to bring the old data into RAM
	void read(void);

	/// called to force an update
	void update(void);

	/// Checks the SimpleTimeout and see if it is currently time for an update.
	void isUpdateTime(void);

	inline void resetTimer(void) { st.start(timeOutToWriteEEPROMData);}

	/// the underlying data... the user can read and write directly from this data
	byte data[maxConfigEEPromByte];

private:
	SimpleTimeout st;
};

#endif // ARDUINO_CONFIGEEPROM_H
