#include "GetConsole.h"
#include <HardwareSerial.h>

void GetConsole::initialize(const unsigned long baudRate)
{
	hs->begin(baudRate);
}

void GetConsole::clearInputBuffer(void)
{
	while(hs->available())
		hs->read();
}

uint8_t GetConsole::getChar()
{
    while (!hs->available());
	return static_cast<uint8_t> ( hs->read() );
}

void GetConsole::getString(char * buffer, const uint8_t length)
{
	clearInputBuffer();
	uint8_t index = 0;
	while (true)
	{
		while (!hs->available());
		char cx = static_cast<char> ( hs->read() );
		if (cx == static_cast<char>(0x7f))  // check for backspace
		{
			if (index)
			{
				hs->write(cx);
				--index;
			}
			continue;
		}
		if (!isprint(cx) ) break;
		*(buffer+index) = cx;
		hs->write(cx);
		++index;
		if (index == static_cast<uint8_t>(length-1) ) break;
	}
	*(buffer+index) = '\0';
}


long GetConsole::getInt(void)
{
	char temp[12];
	getString(temp,12);
	return ::strtol(temp,NULL,0);
}

uint8_t GetConsole::getByte(void)
{
	char temp[8];
	getString(temp,7);
	long ret = ::strtol(temp,NULL,0);
	uint8_t uret = (ret & 0xff);
	return uret;
}


bool GetConsole::getBool(void)
{
	clearInputBuffer();
	while (true)
	{
		int iret = hs->read();
		if (iret == -1) continue;
		char cp = static_cast<char> (iret);
		hs->write(cp);
		switch( cp )
		{
		case 'F':
		case 'f':
		case '0':
		  return false;
		case 't':
		case 'T':
		case '1':
			return true;
		default:
			break;
		}
	}
	return false;
}

unsigned long GetConsole::getIPAdr(void)
{
	int count = 0;
	char str[24];
	getString(str,24);
	uint8_t temp[4];
	char * last;
	const char delimit[] = " .";
	char * cp = ::strtok_r(str,delimit,&last);
	while (count < 4)
	{
		temp[count] = atoi(cp);
		++count;
		cp = ::strtok_r(NULL,delimit,&last);
	}
	unsigned long ul = temp[0];
	for (byte i = 1 ; i < 4 ; ++i)
	{
		ul <<= 8;
		ul += temp[i];
	}
	return  ul;
}
