#ifndef ARDUINO_ICOMDEVICESTATUS_H_
#define  ARDUINO_ICOMDEVICESTATUS_H_

#include <Arduino.h>
#include "IcomDevice.h"

/// Determines the total number of Icom Devices handled by the router.
const byte maxIcomDevices = 10;
/**
* @author James Michener
* Holds all the devices found on the CV
*/

class IcomDeviceStatus
{
public:
	IcomDeviceStatus(void) :
	  maxIcomDevIndex(0)
	  { }
	  ~IcomDeviceStatus(void) {}
	  void dump(void); 
	  /**
	   * @return the index into icomDev for IcomSerial * is
	   * if none found return 0xff  0xff is always an invalid ICOM device ID
       * if is a new source ID... it adds it to table.
	   *  SHOULD ONLY BE CALLED by IcomManager::onMessageWork at the beginning!
	   */
	  byte findDev( IcomSerial * is);

	  // @return the index into icomDev for original device whose ID is x
	  //  if none found returns 0xff  0xff is always an invalid ICOM device ID
	  byte findOrigID(const byte x);

	  // @return the index into icomDev for mapped device whose ID is x
	  //  if none found returns 0xff   0xff is always an invalid ICOM device ID
	  byte findMappedID(const byte x);

	  // @param x  Device ID in question... is it either an original(source) or mapped.
	  bool isIDOrigOrMapped(const uint8_t x);
	  IcomDevice icomDev[maxIcomDevices];
	  byte maxIcomDevIndex;
};


#endif // ARDUINO_ICOMDEVICESTATUS_H_
