#include "IcomManager.h"
#include <HardwareSerial.h>

static const uint8_t myControllerID = 0xe7;
extern const bool bThreePorts;
extern void onNewFreq(unsigned long freq, const byte icomDev);
extern configEEProm ee;
void
IcomManager::checkBroadcast(void)
{
	if (!stBroadcast.isTimedout()) return;
	sendBroadcast(true,false);
}


//@param index is the icomDev index for the source of the message
void IcomManager::makeAndSendBroadcast(IcomSerial * px, const byte index,
	const boolean bFreq, const boolean bMode)
{
	if (index == 0xff)
	{
		return;
	}
	unsigned long freq = icomDev[index].main.freq;

	if (bFreq && freq)
	{
		IcomMessage * im = getMTY();
		im->setSourceDestinationID(icomDev[index].origID,0);
		im->setFreqCmd0(freq);
		px->queueIcomMessage(im);
	}
	byte mode = icomDev[index].main.mode;
	if (bMode && mode != 0xff)
	{
		IcomMessage * im = getMTY();
		im->setSourceDestinationID(icomDev[index].origID,0);
		im->setModeBWCmd1(mode,icomDev[index].main.bandwidth);
		px->queueIcomMessage(im);
	}
}


/// Need to work in the transceive flag in here
void
IcomManager::sendBroadcast(const boolean bFreq, const boolean bMode)
{
	stBroadcast.start(timeOutBroadcast);
	bool bTransceiveFlag = bitRead(this->ConfigurationFlags,ConfigurationFlagBit_Transceive);

	for (byte i = 0 ; i < maxIcomDevIndex ; ++i )
	{
		/// for each radio,  send the frequency,  if the device isn't a radio, there is nothing to send
		if (!bitRead(icomDev[i].typeFlag,ICOMDEVTYPE_RADIOBIT) ) 
			continue;

		if ( &p1 == icomDev[i].comm )  // if this radio is here
		{
			if (bTransceiveFlag || !p2.flagRead(hasRadioOnPort))
				makeAndSendBroadcast(&p2,i,bFreq,bMode);  // send it on the other one
			if (bThreePorts)
			{
				if (bTransceiveFlag || !p3.flagRead(hasRadioOnPort))
					makeAndSendBroadcast(&p3,i,bFreq,bMode);  // send it on the other one
			}
		}
		if ( &p2 == icomDev[i].comm )
		{
			if (bTransceiveFlag || !p1.flagRead(hasRadioOnPort))
				makeAndSendBroadcast(&p1,i,bFreq,bMode);
			if (bThreePorts)
			{
				if (bTransceiveFlag || !p3.flagRead(hasRadioOnPort))
					makeAndSendBroadcast(&p3,i,bFreq,bMode);  // send it on the other one
			}
		}
		if (bThreePorts)
		{
			if ( &p3 == icomDev[i].comm )
			{
				if (bTransceiveFlag || !p1.flagRead(hasRadioOnPort))
					makeAndSendBroadcast(&p1,i,bFreq,bMode);
				if (bTransceiveFlag || !p2.flagRead(hasRadioOnPort))
					makeAndSendBroadcast(&p2,i,bFreq,bMode);  // send it on the other one
			}
		}
	}
}

IcomMessage *
IcomManager::getMTY(void)
{
	IcomMessage * pRet;
	byte iCount = 0;
	while (true)
	{
		pRet = icomMessagePool.getNext();
		if ( pRet->m_Counter )
		{
			++iCount;
			if (iCount < numberOfIcomMesssages) continue;
		    if (!bThreePorts)
			{
				Serial3.println(F("Msg Overflow!"));
			}
			asm(" jmp 0x0000");
//			delay(3000);  // this will cause the watch dog to fire.
		}
		/// mark it as in use.
		pRet->setInUse(); 
		return pRet;
	}
}

void IcomManager::setupConsole(const unsigned long baudRateCIV, const bool bOverRide)
{
//	uint8_t ConfigurationFlags = 0xff;;
	p1.hs = &Serial1;
	p1.imReceive =  getMTY(); // icomMsgPool.get();
	p1.devID = '1';
	p1.flagClear(isDuplexFlag);
	p2.hs = &Serial2;
	p2.imReceive =  getMTY(); // icomMsgPool.get();
	p2.devID = '2';
	p2.flagClear(isDuplexFlag);
	if (bThreePorts )
	{
		p3.hs = & Serial3;
		p3.imReceive = getMTY();
		p3.devID = '3';
		p3.flagClear(isDuplexFlag);
		Serial3.begin(bOverRide ? 19200 : baudRateCIV);
	}
	p0.hs = &Serial;
	p0.imReceive =  getMTY();
	p0.devID = '0';
	p0.flagSet(isDuplexFlag);
	Serial1.begin(baudRateCIV);
	Serial2.begin(baudRateCIV);

}

void IcomManager::setup(const unsigned long baudRateComputer, const unsigned long baudRateCIV, const uint8_t flags , const bool bOverRide)
{
	ConfigurationFlags = flags;
	p1.hs = &Serial1;
	p1.imReceive =  getMTY(); // icomMsgPool.get();
	p1.devID = '1';
	p1.flagClear(isDuplexFlag);
	p2.hs = &Serial2;
	p2.imReceive =  getMTY(); // icomMsgPool.get();
	p2.devID = '2';
	p2.flagClear(isDuplexFlag);
	if (bThreePorts )
	{
		p3.hs = & Serial3;
		p3.imReceive = getMTY();
		p3.devID = '3';
		p3.flagClear(isDuplexFlag);
		Serial3.begin(bOverRide ? 19200 : baudRateCIV);
	}
	else
	{
		Serial3.print(F("CIV baudrate = "));
		Serial3.println(baudRateCIV);
	}
	p0.hs = &Serial;
	p0.imReceive =  getMTY();
	p0.devID = '0';
	p0.flagSet(isDuplexFlag);
	Serial1.begin(baudRateCIV);
	Serial2.begin(baudRateCIV);

	Serial.begin(baudRateComputer);
}

void IcomManager::loop(void)
{
	p1.checkReceive();
	p2.checkReceive();
	p0.checkReceive(); 
	if (bThreePorts)
		p3.checkReceive(); 

    byte llsb = loopCounter & 0x7;
	switch (llsb)
	{
	case 1:
		p1.checkTransmit();
		break;
	case 5:
		p2.checkTransmit();
		break;
    case 7:
        p0.checkTransmit();
        break;
	case 4:
		checkBroadcast();
		break;
	case 3:
		if (bThreePorts)
			p3.checkTransmit();
		break;
	default:
		break;
	}

	++loopCounter;

}

// used only for console operation
void IcomManager::loopTX(void)
{
	p1.checkReceiveThrowAway();
	p2.checkReceiveThrowAway();
//	p0.checkReceiveThrowAway(); 
	if (bThreePorts)
		p3.checkReceiveThrowAway();     
    
    byte llsb = loopCounter & 0x7;
	switch (llsb)
	{
	case 1:
		p1.checkTransmit();
		break;
	case 5:
		p2.checkTransmit();
		break;
	case 3:
		if (bThreePorts)
			p3.checkTransmit();
		break;
	default:
		break;
	}

	++loopCounter;
}

void
IcomManager::onMessageWork(IcomSerial * is)
{
	// this also ads to icomDevices if new source
	byte index = IcomDeviceStatus::findDev(is);  // find the index of the source
	IcomMessage * imp = is->imReceive;
	// if we run out of Icom Devices!
	if (index == 0xff)  // this is never supposed to happen but if it does,
	{
		imp->setMTY();
		return;
	}
	byte desID = imp->getDestinationID();
	uint8_t radioID = ee.data[offsetEEPROMData_ICOMRadioDeviceID];

	// see if the message is a message to the router's controller
	if (radioID != 0xff && desID == myControllerID)
	{
		PowerBandOnMessage(imp);
		return;
	}


	// If full duplex, and configuration bit is set to echo... then echo it back...
	if (is->flagRead(isDuplexFlag) && bitRead(ConfigurationFlags, ConfigurationFlagBit_Echo))
	{
		IcomMessage * newMsg = getMTY();
		newMsg->copy(*imp);
		is->queueIcomMessage(newMsg);
	}

	// if the source has been remapped
	if (IcomDeviceStatus::icomDev[index].mappedID != 0xff)
		imp->setSourceID(IcomDeviceStatus::icomDev[index].mappedID);

	if (index == 0xff)
	{
		imp->setMTY();
		return;  // table is full
	}

	bool bIsNewFreq = false;
	bool sentBroadcastFreqMsg = false;
	// Get frequency and mode from message
	if (imp->isFreq())
	{
		unsigned long newfreq = imp->getFreq() ;

		if (!isControllerID(icomDev[index].origID) )  // is not the controller
		{
			if (bitRead(icomDev[index].typeFlag,ICOMDEVTYPE_VFOSUB) )
				icomDev[index].sub.freq = newfreq  ;
			else
			{
				icomDev[index].main.freq = newfreq  ;
				sendBroadcast(true,false);
				sentBroadcastFreqMsg = true;
			}

			if (!bitRead(icomDev[index].typeFlag,ICOMDEVTYPE_RADIOBIT) )
			{
				bitSet(icomDev[index].typeFlag,ICOMDEVTYPE_RADIOBIT);
				is->flagSet(hasRadioOnPort);
			}
			if (!desID)
			{
				bitSet(icomDev[index].typeFlag,ICOMDEVTYPE_BROADCASTBIT);
				is->flagSet(hasRadioOnPort);
			}
			if (!bitRead(icomDev[index].typeFlag,ICOMDEVTYPE_VFOSUB) )
				onNewFreq(newfreq, icomDev[index].origID);
		}  
		else  // is a controller
			if (imp->isSendOrSetFreqCommand())
				bIsNewFreq = true;
	}  // is a 'normal 'frequency change message

	else if (imp->isFreq25())
	{
		unsigned long newfreq = imp->getFreq25();
		uint8_t vfo = imp->getFreq25VFO();
		if (!isControllerID(icomDev[index].origID))  // is not the controller
		{
			switch (vfo)
			{
			case 0:  // main vfo
				icomDev[index].main.freq = newfreq;
				sendBroadcast(true, false);
				sentBroadcastFreqMsg = true;
				break;
			case 1:  // sub vfo
				icomDev[index].sub.freq = newfreq;
				break;
			default:
				break;
			}

			if (!bitRead(icomDev[index].typeFlag, ICOMDEVTYPE_RADIOBIT))
			{
				bitSet(icomDev[index].typeFlag, ICOMDEVTYPE_RADIOBIT);
				is->flagSet(hasRadioOnPort);
			}
			if (!desID)
			{
				bitSet(icomDev[index].typeFlag, ICOMDEVTYPE_BROADCASTBIT);
				is->flagSet(hasRadioOnPort);
			}
			if (!vfo)  // call external routine.
				onNewFreq(newfreq, icomDev[index].origID);
		}
		else // is a controller  
		{
			bIsNewFreq = true;
		}
	}

	else if (imp->isModeCommand() && !isControllerID(icomDev[index].origID))  // is normal Mode Message?
	{
		uint8_t len = imp->getLen();
		if (bitRead(icomDev[index].typeFlag,ICOMDEVTYPE_VFOSUB) )
		{
			switch (len)
			{
			case 7:
				icomDev[index].sub.bandwidth = imp->getBW();
			case 6:
				icomDev[index].sub.mode = imp->getMode();
				break;
			default:
				break;
			}
		}
		else
		{
			switch (len)
			{
			case 7:
				icomDev[index].main.bandwidth = imp->getBW();
			case 6:
				icomDev[index].main.mode = imp->getMode();
				break;
			default:
				break;
			}
		}
	}
	else if (imp->isMode26Command() && !isControllerID(icomDev[index].origID))
	{
		uint8_t len = imp->getLen();
		uint8_t vfo = imp->getMode26VFO();
		if (vfo)  // Sub VFO
		{
			switch (len)
			{
			case 9:
				icomDev[index].sub.bandwidth = imp->getBW26();
			case 8:
				icomDev[index].sub.dataMode = imp->getDataMode26();
			case 7:
				icomDev[index].sub.mode = imp->getMode26();
			default:
				break;
			}
		}
		else // Main VFO
		{
			switch (len)
			{
			case 9:
				icomDev[index].main.bandwidth = imp->getBW26();
			case 8:
				icomDev[index].main.dataMode = imp->getDataMode26();
			case 7:
				icomDev[index].main.mode = imp->getMode26();
			default:
				break;
			}
		}

	}
	// okay we have a message

	/// destination ID 0 so we have no idea where to send it
	if (!desID) 
	{
		byte cmd = imp->getCommand();
		// kick a broadcast message out from all xcvrs with the latest info
		// to all interested parties except the controller.
		// if we already have sent the message,  don't do it again.
		sendBroadcast(!cmd && !sentBroadcastFreqMsg, cmd == 1);  
        if (bitRead(ConfigurationFlags,ConfigurationFlagBit_TransceiveUSB) )
	    	p0.queueIcomMessage(imp);  // so send it to the controller.
        else
	    	imp->setMTY();  // so throw it away
		return;  ///  don't need to forward a transceiver message
	}

	// has the destination address been mapped somewhere else
	byte destIndex = IcomDeviceStatus::findMappedID(desID);
	if (destIndex != 0xff)  // this is being mapped!
	{
		desID = IcomDeviceStatus::icomDev[desID].mappedID;
		imp->setDestinationID(desID);
//		if (!bThreePorts)
//			Serial3.println(F("dest remap"));
	}
	else
		destIndex = IcomDeviceStatus::findOrigID(desID);

	if (destIndex == 0xff) // can't find destination in table .. send the message everywhere
	{
//		if (isControllerID(desID)  || is == &p0)  //don't speak to the controller unless spoken to
//		{
//			imp->setMTY();
//			return;
//		}
		if ( is == &p0)
		{
			p2.queueIcomMessage(imp);
			copyAndSendMessage(p1,imp);
			if (bThreePorts)
				copyAndSendMessage(p3,imp);
		}
		if ( is == &p1)
		{
			p2.queueIcomMessage(imp);
			copyAndSendMessage(p0,imp);
			if (bThreePorts)
				copyAndSendMessage(p3,imp);
		}
		if ( is == &p2) 
		{
			p1.queueIcomMessage(imp);
			copyAndSendMessage(p0,imp);
			if (bThreePorts)
				copyAndSendMessage(p3,imp);
		}
		if (bThreePorts && is == &p3)
		{
			p1.queueIcomMessage(imp);
			copyAndSendMessage(p2,imp);
			copyAndSendMessage(p0,imp);
		}
		return;
	}

	IcomSerial * isDes = icomDev[destIndex].comm;
	/// same port,  if the other guy is listening he has the message
	if (is == isDes)
	{
		imp->setMTY();
		return;
	}

	/// setting the icomDev VFO SUB bit in the typeFlag
	if (imp->isVFOCommand() && !bitRead(ConfigurationFlags,ConfigurationFlagBit_Subwitch) )
	{
		if (imp->isVFOMainCommand())
		{
			if (bitRead(icomDev[index].typeFlag,ICOMDEVTYPE_RADIOBIT))
				bitClear(icomDev[index].typeFlag,ICOMDEVTYPE_VFOSUB);
			else if (bitRead(icomDev[destIndex].typeFlag,ICOMDEVTYPE_RADIOBIT))
				bitClear(icomDev[destIndex].typeFlag,ICOMDEVTYPE_VFOSUB);
		}
		else if (imp->isVFOSubCommand())
		{
			if (bitRead(icomDev[index].typeFlag,ICOMDEVTYPE_RADIOBIT))
				bitSet(icomDev[index].typeFlag,ICOMDEVTYPE_VFOSUB);
			else if (bitRead(icomDev[destIndex].typeFlag,ICOMDEVTYPE_RADIOBIT))
				bitSet(icomDev[destIndex].typeFlag,ICOMDEVTYPE_VFOSUB);
		}
	}
	// Check to see if it is a command to change frequency 
	// or mode clear the frequency and mode. //imp->isChangeFreqCommand() ||
	if ( imp->isChangeFreqOrModeCommand())
	{
			icomDev[destIndex].sub.freq  = 0;
			icomDev[destIndex].main.freq = 0;
			icomDev[destIndex].sub.mode  = 0xff;
			icomDev[destIndex].main.mode = 0xff;
	}

	//  Check for auto response
	if (isControllerID(imp->getSourceID() ) )
	{
		if (imp->isFreqQuery())
		{
			unsigned long frequency = bitRead(icomDev[destIndex].typeFlag,ICOMDEVTYPE_VFOSUB) ? 
				icomDev[destIndex].sub.freq : icomDev[destIndex].main.freq;
			if (frequency)
			{
				imp->swapSourceDestinationID();
				///  he destID was changed above... need to swap it back
				if (IcomDeviceStatus::icomDev[index].mappedID != 0xFF)
					imp->setDestinationID(IcomDeviceStatus::icomDev[index].origID);
				imp->setFreq(frequency);
				is->queueIcomMessage(imp);
				return;
			}
		}
		else if (imp->isFreq25Query())
		{
			uint8_t vfo = imp->getFreq25VFO();
			unsigned long frequency = vfo ? icomDev[destIndex].sub.freq : icomDev[destIndex].main.freq;
			if (frequency)
			{
				imp->swapSourceDestinationID();
				///  he destID was changed above... need to swap it back
				if (IcomDeviceStatus::icomDev[index].mappedID != 0xFF)
					imp->setDestinationID(IcomDeviceStatus::icomDev[index].origID);
				imp->setFreq25(frequency, vfo);
				is->queueIcomMessage(imp);
				return;
			}

		}
		else if (imp->isModeBWQuery())
		{
			byte mode = bitRead(icomDev[destIndex].typeFlag,ICOMDEVTYPE_VFOSUB) ?  icomDev[destIndex].sub.mode : icomDev[destIndex].main.mode;
			byte bw = bitRead(icomDev[destIndex].typeFlag, ICOMDEVTYPE_VFOSUB) ?
				icomDev[destIndex].sub.bandwidth : icomDev[destIndex].main.bandwidth;
			if (mode != 0xff && bw)
			{
				imp->setModeBW(mode, bw );
				imp->swapSourceDestinationID();
				if (IcomDeviceStatus::icomDev[index].mappedID != 0xFF)
					imp->setDestinationID(IcomDeviceStatus::icomDev[index].origID);
				is->queueIcomMessage(imp);
				return;
			}
		}
	}
	/// Forward the message
	isDes->queueIcomMessage(imp);

	if (bIsNewFreq)
	{
		if (imp->getCommand() == 0x25)
		{
			unsigned long newfreq = imp->getFreq25();
			uint8_t vfo = imp->getFreq25VFO();
			if (vfo)
			{
				icomDev[destIndex].sub.freq = newfreq;
			}
			else
			{
				icomDev[destIndex].main.freq = newfreq;
				sendBroadcast(true, false);
				onNewFreq(newfreq, icomDev[destIndex].origID);
			}
		}
		else {
			unsigned long newfreq = imp->getFreq();
			if (bitRead(icomDev[destIndex].typeFlag, ICOMDEVTYPE_VFOSUB))
				icomDev[destIndex].sub.freq = newfreq;
			else
			{
				icomDev[destIndex].main.freq = newfreq;
				sendBroadcast(true, false);
				onNewFreq(newfreq, icomDev[destIndex].origID);
			}
		}
	}

	if (isControllerID(icomDev[index].origID) && imp->isModeCommand())
	{
		uint8_t len = imp->getLen();
		if (bitRead(icomDev[destIndex].typeFlag, ICOMDEVTYPE_VFOSUB))
		{
			switch (len)
			{
			case 7:
				icomDev[destIndex].sub.bandwidth = imp->getBW();
			case 6:
				icomDev[destIndex].sub.mode = imp->getMode();
				break;
			default:
				break;
			}
		}
		else
		{
			switch (len)
			{
			case 7:
				icomDev[destIndex].main.bandwidth = imp->getBW();
			case 6:
				icomDev[destIndex].main.mode = imp->getMode();
				break;
			default:
				break;
			}
		}
	}
	if (isControllerID(icomDev[index].origID) && imp->isMode26Command())
	{
		uint8_t len = imp->getLen();
		uint8_t vfo = imp->getMode26VFO();
		if (vfo) // sub
		{
			switch (len)
			{
			case 9:
				icomDev[destIndex].sub.bandwidth = imp->getBW26();
			case 8:
				icomDev[destIndex].sub.dataMode = imp->getDataMode26();
			case 7:
				icomDev[destIndex].sub.mode = imp->getMode26();
			default:
				break;
			}
		}
		else
		{
			switch (len)
			{
			case 9:
				icomDev[destIndex].main.bandwidth = imp->getBW26();
			case 8:
				icomDev[destIndex].main.dataMode = imp->getDataMode26();
			case 7:
				icomDev[destIndex].main.mode = imp->getMode26();
			default:
				break;
			}
		}
	}
}


void
	IcomManager::onMessage (IcomSerial * is)
{
	onMessageWork(is);
	/// when we are done processing the message get a new one.
	is->imReceive= getMTY();
}

void 
IcomManager::copyAndSendMessage(IcomSerial & p, IcomMessage * msg)
{
	IcomMessage * newMsg = getMTY();
	newMsg->copy(*msg);
	p.queueIcomMessage(newMsg);
}

void 
IcomManager::sendSplitOffMessage()
{
	uint8_t radioID = ee.data[offsetEEPROMData_ICOMRadioDeviceID];
	uint8_t index = findOrigID(radioID);
	if (index == 0xff)
		return;
	IcomSerial * comm = icomDev[index].comm;
    IcomMessage * im = getMTY();
    im->setSourceDestinationID(myControllerID,radioID);
	im->setSplitOff();
//    im->dump();
    comm->queueIcomMessage(im);
	icomDev[index].typeFlag &= 0xf7; // turn off ICOMDEVTYPE_VFOSPLIT

}

void 
IcomManager::sendMainReadoutMessage()
{
	uint8_t radioID = ee.data[offsetEEPROMData_ICOMRadioDeviceID];
	uint8_t index = findOrigID(radioID);
	if (index == 0xff)
		return;
	IcomSerial * comm = icomDev[index].comm;
    IcomMessage * im = getMTY();
    im->setSourceDestinationID(myControllerID,radioID);
	im->setMainReadout();
//    im->dump();
    comm->queueIcomMessage(im);
	icomDev[index].typeFlag &= 0x7b; // turn off ICOMDEVTYPE_VFOSUB;
}

void 
IcomManager::sendFreqMessage(const unsigned long freq)
{
	uint8_t radioID = ee.data[offsetEEPROMData_ICOMRadioDeviceID];
	uint8_t index = findOrigID(radioID);
	if (index == 0xff)
		return;
	IcomSerial * comm = icomDev[index].comm;
    IcomMessage * im = getMTY();
    im->setSourceDestinationID(myControllerID,radioID);
	im->setFreq(freq);
	im->buffer[3]=5;  //set frequency
    comm->queueIcomMessage(im);
}

// Remember that power messages are done as a test... so there is
// no device table to deal with
void IcomManager::sendPowerMessage(const byte value)
{
	uint8_t radioID = ee.data[offsetEEPROMData_ICOMRadioDeviceID];
	if (radioID >= 0xe0)
		return;

    IcomMessage * im = getMTY();
    im->setSourceDestinationID(myControllerID,radioID);
    im->setPower(value);
	uint8_t index = findOrigID(radioID);
	if (index == 0xff)  // do not know port of radio  ... send it everywhere
	{
		p1.queueIcomMessage(im);
		copyAndSendMessage(p2,im);
		if (bThreePorts)
			copyAndSendMessage(p3,im);
	}
	else // know where to send it..
		icomDev[index].comm->queueIcomMessage(im);
}
void IcomManager::sendPowerQueryMessge()
{
 	uint8_t radioID = ee.data[offsetEEPROMData_ICOMRadioDeviceID];
	if (radioID >= 0xe0)
		return;
	IcomMessage * im = getMTY();
    im->setSourceDestinationID(myControllerID,radioID);
    im->setPowerQuery();
	uint8_t index = findOrigID(radioID);
	if (index == 0xff)  // do not know port of radio ... send it everywhere
	{
	    p1.queueIcomMessage(im);
		copyAndSendMessage(p2,im);
		if (bThreePorts)
			copyAndSendMessage(p3,im);
	}
	else  // know where to send it..
		icomDev[index].comm->queueIcomMessage(im);
}
